﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Runtime.InteropServices;
using System.Security.Cryptography;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;

namespace FaxToBinary
{
	public class Project : ViewModelBase
	{
		public Project()
		{
			LoadSettingsFromPath(Path.Combine(Directory.GetCurrentDirectory(), "last_settings.txt"));
		}

		private BitmapSource[] _images = null;
		public BitmapSource[] Images
		{
			get { return _images; }
			private set
			{
				SetProperty(ref _images, value, (oldValue, newValue) =>
				{
					OnPropertyChanged(nameof(IsValid));
					OnPropertyChanged(nameof(MaximumIndex));
				});
			}
		}

		public bool IsValid { get { return Images != null && Images.Length > 0; } }
		public int MaximumIndex { get { return Images != null ? Images.Length - 1 : -1; } }

		public Dictionary<string, char> Glyphs { get; private set; } = new Dictionary<string, char>();

		private string[] _results = null;
		public string[] Results
		{
			get { return _results; }
			private set { SetProperty(ref _results, value); }
		}

		private bool _loadedSettingsForImage = false;
		private bool _loadInProgress = false;

		private string _filePath = "";
		public string FilePath
		{
			get { return _filePath; }
			private set
			{
				SetProperty(ref _filePath, value, (oldValue, newValue) =>
				{
					OnPropertyChanged(nameof(Identifier));
					OnPropertyChanged(nameof(FileRoot));
				});
			}
		}

		public string FileRoot
		{
			get { return Path.Combine(Directory.GetCurrentDirectory(), Identifier); }
		}

		[ThreadStatic]
		private static MD5 _md5;
		public static string CreateHash(byte[] text)
		{
			if (_md5 == null)
			{
				_md5 = MD5.Create();
			}
			var hashBytes = _md5.ComputeHash(text);
			var sb = new StringBuilder();
			for (var i = 0; i < hashBytes.Length; i++)
			{
				sb.Append(hashBytes[i].ToString("X2"));
			}
			return sb.ToString();
		}

		public string Identifier
		{
			get { return CreateHash(Encoding.ASCII.GetBytes(FilePath)); }
		}
		
		private BitmapSource _activeImageSource = null;
		public BitmapSource ActiveImageSource
		{
			get { return _activeImageSource; }
			private set
			{
				SetProperty(ref _activeImageSource, value, (oldValue, newValue) =>
				{
					OnPropertyChanged(nameof(CharactersPerLine));
					OnPropertyChanged(nameof(Lines));
					UpdatePreviewImageSource();
				});
			}
		}
		
		private int _activeIndex = -1;
		public int ActiveIndex
		{
			get { return _activeIndex; }
			set
			{
				SetProperty(ref _activeIndex, value, (oldValue, newValue) =>
				{
					ActiveImageSource = (Images != null && _activeIndex >= 0 && _activeIndex < Images.Length) ? Images[_activeIndex] : null;
				});
			}
		}

		private BitmapSource _activePreviewImageSource = null;
		public BitmapSource ActivePreviewImageSource
		{
			get { return _activePreviewImageSource; }
			private set { SetProperty(ref _activePreviewImageSource, value); }
		}

		private byte[] GetCharacterPixelData(BitmapSource image, int characterIndex, int lineIndex, System.Windows.Media.PixelFormat outputFormat)
		{
			var characterOffsetX = OffsetX + characterIndex * CharacterWidth;
			var characterOffsetY = OffsetY + lineIndex * LineHeight + (FixLineHeightRoundingError ? lineIndex / 2 : 0);
			var width = CharacterWidth;
			var height = LineHeight;

			var stride = width * ((outputFormat.BitsPerPixel + 7) / 8);
			var pixelData = new byte[stride * height];
			image.CopyPixels(new Int32Rect(characterOffsetX, characterOffsetY, width, height), pixelData, stride, 0);
			return pixelData;
		}

		private void UpdatePreviewImageSource()
		{
			ActivePreviewImageSource = null;

			if (ActiveImageSource != null &&
				ActiveCharacterPreview >= 0 && ActiveCharacterPreview < CharactersPerLine &&
				ActiveLinePreview >= 0 && ActiveLinePreview < Lines)
			{
				var width = CharacterWidth;
				var height = LineHeight;
				var outputFormat = PixelFormats.Bgr24;
				var pixelData = GetCharacterPixelData(ActiveImageSource, ActiveCharacterPreview, ActiveLinePreview, outputFormat);
				var stride = width * ((outputFormat.BitsPerPixel + 7) / 8);
				ActivePreviewImageSource = BitmapSource.Create(width, height, 96, 96, outputFormat, null, pixelData, stride);
			}
		}

		private int _activeCharacterPreview = 0;
		public int ActiveCharacterPreview
		{
			get { return _activeCharacterPreview; }
			set
			{
				SetProperty(ref _activeCharacterPreview, value, (oldValue, newValue) =>
				{
					UpdatePreviewImageSource();
				});
			}
		}

		private int _activeLinePreview = 0;
		public int ActiveLinePreview
		{
			get { return _activeLinePreview; }
			set
			{
				SetProperty(ref _activeLinePreview, value, (oldValue, newValue) =>
				{
					UpdatePreviewImageSource();
				});
			}
		}

		private int _startIndex = 0;
		public int StartIndex
		{
			get { return _startIndex; }
			set
			{
				SetProperty(ref _startIndex, value, (oldValue, newValue) =>
				{
					SaveSettings();
				});
			}
		}

		private int _endIndex = 0;
		public int EndIndex
		{
			get { return _endIndex; }
			set
			{
				SetProperty(ref _endIndex, value, (oldValue, newValue) =>
				{
					SaveSettings();
				});
			}
		}

		private int _characterWidth = 20;
		public int CharacterWidth
		{
			get { return _characterWidth; }
			set
			{
				SetProperty(ref _characterWidth, value, (oldValue, newValue) =>
				{
					OnPropertyChanged(nameof(CharactersPerLine));
					UpdatePreviewImageSource();
					SaveSettings();
				});
			}
		}

		public int CharactersPerLine
		{
			get { return ActiveImageSource != null ? ((int)ActiveImageSource.Width - OffsetX) / CharacterWidth : 0; }
		}

		private int _lineHeight = 24;
		public int LineHeight
		{
			get { return _lineHeight; }
			set
			{
				SetProperty(ref _lineHeight, value, (oldValue, newValue) =>
				{
					OnPropertyChanged(nameof(Lines));
					UpdatePreviewImageSource();
					SaveSettings();
				});
			}
		}

		public int Lines
		{
			get
			{
				var lineHeight = LineHeight + (FixLineHeightRoundingError ? 0.5f : 0.0f);
				return ActiveImageSource != null ? (int)(((int)ActiveImageSource.Height - OffsetY) / lineHeight) : 0;
			}
		}

		private int _offsetX = 0;
		public int OffsetX
		{
			get { return _offsetX; }
			set
			{
				SetProperty(ref _offsetX, value, (oldValue, newValue) =>
				{
					OnPropertyChanged(nameof(CharactersPerLine));
					UpdatePreviewImageSource();
					SaveSettings();
				});
			}
		}

		private int _offsetY = 0;
		public int OffsetY
		{
			get { return _offsetY; }
			set
			{
				SetProperty(ref _offsetY, value, (oldValue, newValue) =>
				{
					OnPropertyChanged(nameof(Lines));
					UpdatePreviewImageSource();
					SaveSettings();
				});
			}
		}

		private bool _fixLineHeightRoundingError = false;
		public bool FixLineHeightRoundingError
		{
			get { return _fixLineHeightRoundingError; }
			set
			{
				SetProperty(ref _fixLineHeightRoundingError, value, (oldValue, newValue) =>
				{
					UpdatePreviewImageSource();
					SaveSettings();
				});
			}
		}
		
		public void Reset()
		{
			_loadedSettingsForImage = false;
			Images = null;
			Glyphs.Clear();
			Results = null;
			ActiveIndex = -1;
			FilePath = "";
		}

		public void Load(string filePath)
		{
			FilePath = filePath;
			Directory.CreateDirectory(FileRoot);

			Results = null;

			var source = (Bitmap)Image.FromFile(FilePath);
			var images = new BitmapSource[source.GetFrameCount(FrameDimension.Page)];
			Parallel.For(0, images.Length, (i) =>
			{
				var bitmap = (Bitmap)Image.FromFile(FilePath);
				bitmap.SelectActiveFrame(FrameDimension.Page, i);

				var width = bitmap.Width;
				var height = bitmap.Height;
				var data = bitmap.LockBits(new Rectangle(0, 0, width, height), ImageLockMode.ReadWrite, System.Drawing.Imaging.PixelFormat.Format24bppRgb);
				var pixelData = new byte[data.Stride * height];
				Marshal.Copy(data.Scan0, pixelData, 0, pixelData.Length);
				bitmap.UnlockBits(data);

				var outputformat = PixelFormats.Bgr24;
				var stride = width * ((outputformat.BitsPerPixel + 7) / 8);
				images[i] = BitmapSource.Create(width, height, 96, 96, outputformat, null, pixelData, stride);
				images[i].Freeze();
			});
			Images = images;

			EndIndex = MaximumIndex;

			LoadSettings();
			LoadGlyphs();

			ActiveIndex = StartIndex;
		}

		private static byte[] StringToBytes(string hexString)
		{
			var bytes = new byte[hexString.Length / 2];
			for (var i = 0; i < hexString.Length - 1; i += 2)
			{
				Func<int, int> GetHexValue = (hex) =>
				{
					if (hex >= '0' && hex <= '9')
					{
						return hex - '0';
					}
					if (hex >= 'a' && hex <= 'f')
					{
						return hex - 'a' + 10;
					}
					throw new Exception("Unexpected value in hex string.");
				};
				var msb = GetHexValue(hexString[i]);
				var lsb = GetHexValue(hexString[i + 1]);
				bytes[i / 2] = (byte)((msb << 4) + lsb);
			}
			return bytes;
		}

		public void Save(string filePath)
		{
			if (Results != null)
			{
				var sb = new StringBuilder();
				for (var i = StartIndex; i <= EndIndex; ++i)
				{
					sb.Append(Results[i - StartIndex]);
				}
				var result = StringToBytes(sb.ToString());
				File.WriteAllBytes(filePath, result);
			}
		}

		public async Task Process()
		{
			if (IsValid)
			{
				var results = new string[EndIndex - StartIndex + 1];

				var waitLock = new object();
				var tokenSource = new CancellationTokenSource();
				var token = tokenSource.Token;
				var options = new ParallelOptions { CancellationToken = token };

				var tasks = new Task[results.Length];
				for (var i = StartIndex; i <= EndIndex && !token.IsCancellationRequested; ++i)
				{
					int taskIndex = i - StartIndex;
					int imageIndex = i;
					tasks[taskIndex] = Task.Run(() =>
					{
						var sb = new StringBuilder();

						for (var line = 0; line < Lines && !token.IsCancellationRequested; ++line)
						{
							for (var characterIndex = 0; characterIndex < CharactersPerLine && !token.IsCancellationRequested; ++characterIndex)
							{
								var width = CharacterWidth;
								var height = LineHeight;
								var outputFormat = PixelFormats.Bgr24;
								var pixelData = GetCharacterPixelData(Images[imageIndex], characterIndex, line, outputFormat);
								var hash = CreateHash(pixelData);

								lock (waitLock)
								{
									if (!token.IsCancellationRequested)
									{
										if (Glyphs.TryGetValue(hash, out char character))
										{
											if (character != ' ')
											{
												sb.Append(character);
											}
										}
										else
										{
											Application.Current.Dispatcher.Invoke(() =>
											{
												var dialog = new GlyphInputDialog();
												var stride = width * ((outputFormat.BitsPerPixel + 7) / 8);
												dialog.Image = BitmapSource.Create(width, height, 96, 96, outputFormat, null, pixelData, stride);
												if (dialog.ShowDialog() == true)
												{
													character = dialog.Text.Length == 1 ? char.ToLower(dialog.Text[0]) : ' ';
													Glyphs[hash] = character;
													if (character != ' ')
													{
														sb.Append(character);
													}
													AddGlyph(hash, character);
												}
												else
												{
													tokenSource.Cancel();
												}
											});
										}
									}
								}
							}
						}

						results[taskIndex] = sb.ToString();
					}, token);
				}

				try
				{
					await Task.WhenAll(tasks);
				}
				catch
				{
				}

				if (token.IsCancellationRequested)
				{
					MessageBox.Show("Processing failed.", "Process", MessageBoxButton.OK, MessageBoxImage.Error);
				}
				else
				{
					Results = results;
					MessageBox.Show("Processing complete.", "Process", MessageBoxButton.OK, MessageBoxImage.Information);
				}
			}
		}

		private void SaveSettingsToPath(string path)
		{
			Directory.CreateDirectory(Path.GetDirectoryName(path));
			File.WriteAllLines(path, new List<string>
			{
				StartIndex.ToString(),
				EndIndex.ToString(),
				CharacterWidth.ToString(),
				LineHeight.ToString(),
				OffsetX.ToString(),
				OffsetY.ToString(),
				FixLineHeightRoundingError.ToString(),
			});
		}

		private void LoadSettingsFromPath(string path)
		{
			if (File.Exists(path))
			{
				_loadInProgress = true;
				try
				{
					var lines = File.ReadAllLines(path);
					StartIndex = int.Parse(lines[0]);
					EndIndex = int.Parse(lines[1]);
					CharacterWidth = int.Parse(lines[2]);
					LineHeight = int.Parse(lines[3]);
					OffsetX = int.Parse(lines[4]);
					OffsetY = int.Parse(lines[5]);
					FixLineHeightRoundingError = lines.Length >= 7 && bool.Parse(lines[6]);
				}
				finally
				{
					_loadInProgress = false;
				}
			}
		}

		private void SaveSettings()
		{
			if (!_loadInProgress)
			{
				if (_loadedSettingsForImage)
				{
					SaveSettingsToPath(Path.Combine(FileRoot, "settings.txt"));
				}
				SaveSettingsToPath(Path.Combine(Directory.GetCurrentDirectory(), "last_settings.txt"));
			}
		}

		private void LoadSettings()
		{
			LoadSettingsFromPath(Path.Combine(FileRoot, "settings.txt"));
			_loadedSettingsForImage = true;
		}
		
		private void AddGlyph(string hash, char character)
		{
			Directory.CreateDirectory(FileRoot);
			var glyphPath = Path.Combine(FileRoot, string.Format("glyph_{0}.txt", hash));
			File.WriteAllText(glyphPath, character.ToString());
		}

		private void LoadGlyphs()
		{
			const string GlyphPrefix = "glyph_";
			if (Directory.Exists(FileRoot))
			{
				foreach (var glyphPath in Directory.EnumerateFiles(FileRoot, GlyphPrefix + "*.txt", SearchOption.TopDirectoryOnly))
				{
					var name = Path.GetFileNameWithoutExtension(glyphPath).Substring(GlyphPrefix.Length);
					Glyphs[name] = File.ReadAllText(glyphPath)[0];
				}
			}
		}
	}
}
